use crate::persistence::{SealError, SealedFile};
use serde::{Deserialize, Serialize};
use serde_with::{serde_as, TimestampSeconds};
use std::collections::{HashMap, HashSet};
use std::sync::RwLock;
use std::time::{Duration, SystemTime, UNIX_EPOCH};

type IP = String;
const LOCAL_INFO_FILE: &str = "/host/main/node_info";

#[serde_as]
#[derive(Clone, Serialize, Deserialize, PartialEq, Eq, Ord, PartialOrd)]
pub struct NodeInfo {
    #[serde_as(as = "TimestampSeconds")]
    pub started_at: SystemTime,
    #[serde_as(as = "TimestampSeconds")]
    pub keepalive: SystemTime,
    pub mint_requests: u64,
    pub mints: u64,
    pub mratls_conns: u64,
    pub net_attacks: u64,
    pub public: bool,
    pub restarts: u64,
    pub disk_attacks: u64,
}

impl NodeInfo {
    pub fn new_empty() -> Self {
        NodeInfo {
            started_at: SystemTime::now(),
            keepalive: SystemTime::now(),
            mint_requests: 0,
            mints: 0,
            mratls_conns: 0,
            net_attacks: 0,
            public: false,
            restarts: 0,
            disk_attacks: 0,
        }
    }

    pub fn is_newer_than(&self, older_self: &Self) -> bool {
        self.keepalive > older_self.keepalive
    }

    pub fn to_json(&self) -> String {
        serde_json::to_string(self).unwrap() // can fail only if time goes backwards :D
    }

    pub fn to_metrics(&self, ip: &str) -> String {
        let started_at = self.started_at.duration_since(UNIX_EPOCH).unwrap_or(Duration::ZERO);
        let keepalive = self.keepalive.duration_since(UNIX_EPOCH).unwrap_or(Duration::ZERO);
        let labels = format!("{{ip=\"{}\", public=\"{}\"}}", ip, self.public);

        let mut res = String::new();
        res.push_str(&format!("started_at{} {}\n", labels, started_at.as_secs()));
        res.push_str(&format!("keepalive{} {}\n", labels, keepalive.as_secs()));
        res.push_str(&format!("mint_requests{} {}\n", labels, self.mint_requests));
        res.push_str(&format!("mints{} {}\n", labels, self.mints));
        res.push_str(&format!("mratls_conns{} {}\n", labels, self.mratls_conns));
        res.push_str(&format!("net_attacks{} {}\n", labels, self.net_attacks));
        res.push_str(&format!("restarts{} {}\n", labels, self.restarts));
        res.push_str(&format!("disk_attacks{} {}\n", labels, self.disk_attacks));
        res
    }

    pub fn load() -> Self {
        match Self::read(LOCAL_INFO_FILE) {
            Ok(mut info) => {
                info.mratls_conns = 0;
                info.restarts += 1;
                info
            }
            Err(SealError::Attack(e)) => {
                println!("The local node file is corrupted: {}", e);
                let mut info = Self::new_empty();
                info.disk_attacks += 1; // add very first disk attack
                info
            }
            Err(_) => Self::new_empty(),
        }
    }

    pub fn save(&self) {
        if let Err(e) = self.write(LOCAL_INFO_FILE) {
            println!("Could not save node info: {}", e);
        }
    }
}

/// Multithreaded state, designed to be
/// shared everywhere in the code
pub struct State {
    my_ip: String,
    nodes: RwLock<HashMap<IP, NodeInfo>>,
    conns: RwLock<HashSet<IP>>,
}

impl State {
    pub fn new(my_ip: String) -> Self {
        let mut nodes = HashMap::new();
        let my_info = NodeInfo::load();
        nodes.insert(my_ip.clone(), my_info);
        Self { my_ip, nodes: RwLock::new(nodes), conns: RwLock::new(HashSet::new()) }
    }

    pub fn add_conn(&self, ip: &str) {
        self.increase_mratls_conns();
        self.add_mratls_conn(ip);
    }

    pub fn delete_conn(&self, ip: &str) {
        self.decrease_mratls_conns();
        self.delete_mratls_conn(ip);
    }

    fn add_mratls_conn(&self, ip: &str) {
        if let Ok(mut conns) = self.conns.write() {
            conns.insert(ip.to_string());
        }
    }

    fn delete_mratls_conn(&self, ip: &str) {
        if let Ok(mut conns) = self.conns.write() {
            conns.remove(ip);
        }
    }

    pub fn increase_mint_requests(&self) {
        if let Ok(mut nodes) = self.nodes.write() {
            if let Some(my_info) = nodes.get_mut(&self.my_ip) {
                *my_info = NodeInfo { mint_requests: my_info.mint_requests + 1, ..my_info.clone() };
            }
        }
    }

    pub fn increase_mints(&self) {
        if let Ok(mut nodes) = self.nodes.write() {
            if let Some(my_info) = nodes.get_mut(&self.my_ip) {
                *my_info = NodeInfo { mints: my_info.mints + 1, ..my_info.clone() };
            }
        }
    }

    pub fn increase_mratls_conns(&self) {
        if let Ok(mut nodes) = self.nodes.write() {
            if let Some(my_info) = nodes.get_mut(&self.my_ip) {
                *my_info = NodeInfo { mratls_conns: my_info.mratls_conns + 1, ..my_info.clone() };
            }
        }
    }

    pub fn decrease_mratls_conns(&self) {
        if let Ok(mut nodes) = self.nodes.write() {
            if let Some(my_info) = nodes.get_mut(&self.my_ip) {
                if my_info.mratls_conns > 0 {
                    *my_info =
                        NodeInfo { mratls_conns: my_info.mratls_conns - 1, ..my_info.clone() };
                }
            }
        }
    }

    pub fn increase_disk_attacks(&self) {
        if let Ok(mut nodes) = self.nodes.write() {
            if let Some(my_info) = nodes.get_mut(&self.my_ip) {
                *my_info = NodeInfo { disk_attacks: my_info.disk_attacks + 1, ..my_info.clone() };
            }
        }
    }

    pub fn increase_net_attacks(&self) {
        if let Ok(mut nodes) = self.nodes.write() {
            if let Some(my_info) = nodes.get_mut(&self.my_ip) {
                *my_info = NodeInfo { net_attacks: my_info.net_attacks + 1, ..my_info.clone() };
            }
        }
    }

    pub fn declare_myself_public(&self) {
        if let Ok(mut nodes) = self.nodes.write() {
            if let Some(my_info) = nodes.get_mut(&self.my_ip) {
                *my_info = NodeInfo { public: true, ..my_info.clone() };
            }
        }
    }

    pub fn get_nodes(&self) -> Vec<(String, NodeInfo)> {
        if let Ok(nodes) = self.nodes.read() {
            return nodes.iter().map(|(k, v)| (k.clone(), v.clone())).collect();
        }
        Vec::new()
    }

    pub fn get_my_ip(&self) -> String {
        self.my_ip.clone()
    }

    pub fn get_my_info(&self) -> NodeInfo {
        if let Ok(nodes) = self.nodes.read() {
            if let Some(found_info) = nodes.get(&self.my_ip) {
                return found_info.clone();
            }
        }
        NodeInfo::new_empty()
    }

    pub fn get_connected_ips(&self) -> Vec<String> {
        if let Ok(conns) = self.conns.read() {
            return conns.iter().map(|n| n.clone()).collect();
        }
        Vec::new()
    }

    // returns a random node that does not have an active connection
    pub fn get_random_disconnected_ip(&self) -> Option<String> {
        use rand::{rngs::OsRng, RngCore};

        let conn_ips = self.get_connected_ips();
        if let Ok(nodes) = self.nodes.read() {
            let skip = OsRng.next_u64().try_into().unwrap_or(0) % nodes.len();
            return nodes
                .keys()
                .map(|ip| ip.to_string())
                .filter(|ip| ip != &self.my_ip && !conn_ips.contains(ip))
                .cycle()
                .nth(skip);
        }
        None
    }

    /// This returns true if the update should be further forwarded
    /// For example, we never forward our own updates that came back
    pub fn process_node_update(&self, (node_ip, node_info): (String, NodeInfo)) -> bool {
        let is_update_mine = node_ip.eq(&self.my_ip);
        let mut is_update_new = false;
        if let Ok(mut nodes) = self.nodes.write() {
            is_update_new = nodes
                .get(&node_ip)
                .map(|curr_info| node_info.is_newer_than(&curr_info))
                .unwrap_or(true);
            if is_update_new {
                let _ = nodes.insert(node_ip.clone(), node_info.clone());
            }
        }
        if is_update_new {
            println!("Inserting: {}, {}", node_ip, node_info.to_json());
        }
        is_update_new && !is_update_mine
    }

    pub fn remove_inactive_nodes(&self, max_age: u64) {
        if let Ok(mut nodes) = self.nodes.write() {
            // TODO: Check if it is possible to corrupt SGX system time
            let now = SystemTime::now();
            nodes.retain(|_, n| {
                let age = now.duration_since(n.keepalive).unwrap_or(Duration::ZERO).as_secs();
                age <= max_age
            });
        }
    }
}
